/* gvSIG Mini. A free mobile phone viewer of free maps.
 *
 * Copyright (C) 2010 Prodevelop.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 *   Prodevelop, S.L.
 *   Pza. Don Juan de Villarrasa, 14 - 5
 *   46001 Valencia
 *   Spain
 *
 *   +34 963 510 612
 *   +34 963 510 968
 *   prode@prodevelop.es
 *   http://www.prodevelop.es
 *
 *   gvSIG Mini has been partially funded by IMPIVA (Instituto de la Peque�a y
 *   Mediana Empresa de la Comunidad Valenciana) &
 *   European Union FEDER funds.
 *   
 *   2010.
 *   author Alberto Romeu aromeu@prodevelop.es
 */

package es.prodevelop.gvsig.mini.common;

import java.io.IOException;
import java.io.InputStream;

/**
 * A common interface to abstract methods of different platforms (J2SE, CLDC and
 * Android)
 * 
 * @author aromeu
 * 
 */
public interface IContext {
	
	/**
	 * TileGenerator has been canceled
	 * @see TileGenerator
	 * @see Cancellable
	 */
	public static final int DOWNLOAD_CANCELED = 0;
	
	/**
	 * TileGenerator has finished to download tiles
	 * @see TileGenerator
	 */
	public static final int FINISH_DOWNLOAD = DOWNLOAD_CANCELED + 1;
	
	/**
	 * TileGenerator has started to download tiles
	 * @see TileGenerator
	 */
	public static final int START_DOWNLOAD = FINISH_DOWNLOAD + 1;
	
	/**
	 * A single tile has been downloaded
	 * @see TileGenerator
	 * @see TileProvider
	 */
	public static final int TILE_DOWNLOADED = START_DOWNLOAD + 1;
	
	/**
	 * The total number of tiles to be downloaded has been retrieved
	 * @see TileGenerator
	 */
	public static final int TOTAL_TILES_COUNT = TILE_DOWNLOADED + 1;	

	/**
	 * Opens a file embedded in the assets folder (in Android)
	 * 
	 * @param fileName
	 *            The name of the file
	 * @return An InputStream to the file
	 */
	public InputStream openAssetFile(String fileName) throws IOException;

	/**
	 * Returns the path for base layer file to load it
	 * 
	 * @return
	 */
	public String getBaseLayerFilePath();

	/**
	 * Returns the path for base layer file to persist it
	 * 
	 * @return
	 */
	public String getBaseLayerPersistFilePath();

	/**
	 * Usually /sdcard
	 * This path is used when storing tiles on the file system
	 * 
	 */
	public String getExternalStorageDirectoryPath();
	
	
	/**
	 * This path is used when storing tiles on the file system
	 * @param path
	 * @return
	 */
	public void setExternalStorageDirectoryPath(String path);

	/**
	 * Creates a IBitmap from a byte array
	 * 
	 * @param data
	 * @param offset
	 * @param length
	 * @return
	 */
	public IBitmap decodeByteArray(byte[] data, int offset, int length) throws IOException;
	
	/**
	 * Creates a IBitmap from an InputStream
	 * 
	 * @param in
	 * @return
	 */
	public IBitmap decodeByteArray(InputStream in) throws IOException;
	
	/**
	 * Creates a IBitmap from a file name
	 * 
	 * @param fn
	 * @return
	 */
	public IBitmap decodeByteArray(String filePath) throws IOException;

	/**
	 * Sends a message to the IHandler. For Android this method needs to call
	 * obtainMessage on the Handler, and sendToTarget
	 * 
	 * @param handler
	 *            An IHandler
	 * @param event
	 *            The event to send. Use getMessage() to get a String message of the event
	 * @param ID
	 *            A Constant to send with the message
	 */
	public void sendMessage(IHandler handler, IEvent event, int ID);

	/**
	 * Sends an empty message to the IHandler. For Android this method needs to
	 * call sendEmptyMessage
	 * 
	 * @param handler
	 *            An IHandler
	 * @param ID
	 *            A Constant to send with the message
	 */
	public void sendMessage(IHandler handler, int ID);

	/**
	 * Looks for the Bitmap with the ID in the ResourceLoader
	 * 
	 * @param ID
	 *            The id of the bitmap -> R.drawable.ID
	 * @return a Bitmap
	 */
	public IBitmap getBitmapFromResource(int ID);

	/**
	 * True if the SD card is mounted.
	 * 
	 * @return
	 */
	public boolean isSDMounted();

	/**
	 * This method should instantiate a new IRect instance. The IRect
	 * implementation is dependant of the platform (Android, CLDC, J2SE...)
	 * 
	 * @return
	 */
	public IRect getRectangle();
	
	/**
	 * The ID of the resource for the tile that represents that a tile is being loaded
	 * @return
	 */
	public int getDefaultLoadingTileID(); 
	
	/**
	 * The ID of the resource for the tile that represents that a tile is not exists when in offline mode
	 * @return
	 */
	public int getDefaultOfflineTileID();
	
	/**
	 * Creates a new IBitmap instance (depending on the implementation)
	 * @param width The width in pixels of the bitmap
	 * @param height The height in pixels of the bitmap
	 * @return
	 */
	public IBitmap createBitmap(int width, int height);
}
