package es.prodevelop.gvsig.mini.common;

import java.security.SecureRandom;

import javax.crypto.Cipher;
import javax.crypto.KeyGenerator;
import javax.crypto.SecretKey;
import javax.crypto.spec.SecretKeySpec;

import es.prodevelop.gvsig.mini.exceptions.BaseException;

public abstract class IEncryptor {

	public abstract String getUniqueID();

	public abstract String getIMEI();

	public static final String OFF = "_OFFF_";
	public static final String DEV_ID = "_DEVID_";

	/**
	 * Checks if the text starts either witt OFF or DEV_ID and decrypts it using
	 * either getIMEI() or getUniqueID()
	 * 
	 * @param clearText
	 * @return
	 * @throws Exception
	 */
	public String tryDecrypt(String clearText) throws Exception {
		String res = "";
		if (clearText.startsWith(OFF)) {
			res = clearText.substring(IEncryptor.OFF.length(),
					clearText.length());
			res = decrypt(res, getIMEI());
			return res;
		} else if (clearText.startsWith(DEV_ID)) {
			res = clearText.substring(IEncryptor.DEV_ID.length(),
					clearText.length());
			res = decrypt(res, getUniqueID());
			return res;
		} else
			throw new BaseException("Not decrypted");
	}

	/**
	 * Encrypts using the key given by getUniqueID
	 * 
	 * @param cleartext
	 *            The text to encrypt
	 * @return The encrypted text
	 * @throws Exception
	 */
	public String encrypt(String cleartext) throws Exception {
		return encrypt(cleartext, getUniqueID());
	}

	/**
	 * Decrypts some text using the key given by getUniqueID
	 * 
	 * @param encrypted
	 *            The encrypted text
	 * @return The decrypted text
	 * @throws Exception
	 */
	public String decrypt(String encrypted) throws Exception {
		return decrypt(encrypted, getUniqueID());
	}

	/**
	 * Encrypts some text
	 * 
	 * @param clearText
	 *            The text to encrypt
	 * @param key
	 *            The key to use
	 * @return The encrypted text
	 * @throws Exception
	 */
	public String encrypt(String clearText, String key) throws Exception {
		byte[] rawKey = getRawKey(key.getBytes());
		byte[] result = encrypt(rawKey, clearText.getBytes());
		return toHex(result);
	}

	/**
	 * Decrypts some text
	 * 
	 * @param encryptedText
	 *            The encrypted text
	 * @param key
	 *            The key in which the text was encrypted
	 * @return The decrypted text
	 * @throws Exception
	 */
	public String decrypt(String encryptedText, String key) throws Exception {
		byte[] rawKey = getRawKey(key.getBytes());
		byte[] enc = toByte(encryptedText);
		byte[] result = decrypt(rawKey, enc);
		return new String(result);
	}

	private byte[] getRawKey(byte[] seed) throws Exception {
		KeyGenerator kgen = KeyGenerator.getInstance("AES");
		SecureRandom sr = SecureRandom.getInstance("SHA1PRNG");
		sr.setSeed(seed);
		kgen.init(128, sr); // 192 and 256 bits may not be available
		SecretKey skey = kgen.generateKey();
		byte[] raw = skey.getEncoded();
		return raw;
	}

	private byte[] encrypt(byte[] raw, byte[] clear) throws Exception {
		SecretKeySpec skeySpec = new SecretKeySpec(raw, "AES");
		Cipher cipher = Cipher.getInstance("AES");
		cipher.init(Cipher.ENCRYPT_MODE, skeySpec);
		byte[] encrypted = cipher.doFinal(clear);
		return encrypted;
	}

	private byte[] decrypt(byte[] raw, byte[] encrypted) throws Exception {
		SecretKeySpec skeySpec = new SecretKeySpec(raw, "AES");
		Cipher cipher = Cipher.getInstance("AES");
		cipher.init(Cipher.DECRYPT_MODE, skeySpec);
		byte[] decrypted = cipher.doFinal(encrypted);
		return decrypted;
	}

	public String toHex(String txt) {
		return toHex(txt.getBytes());
	}

	public String fromHex(String hex) {
		return new String(toByte(hex));
	}

	public byte[] toByte(String hexString) {
		int len = hexString.length() / 2;
		byte[] result = new byte[len];
		for (int i = 0; i < len; i++)
			result[i] = Integer.valueOf(hexString.substring(2 * i, 2 * i + 2),
					16).byteValue();
		return result;
	}

	public String toHex(byte[] buf) {
		if (buf == null)
			return "";
		StringBuffer result = new StringBuffer(2 * buf.length);
		for (int i = 0; i < buf.length; i++) {
			appendHex(result, buf[i]);
		}
		return result.toString();
	}

	private final static String HEX = "0123456789ABCDEF";

	private void appendHex(StringBuffer sb, byte b) {
		sb.append(HEX.charAt((b >> 4) & 0x0f)).append(HEX.charAt(b & 0x0f));
	}

}
