/* gvSIG. Geographic Information System of the Valencian Government
*
* Copyright (C) 2007-2008 Infrastructures and Transports Department
* of the Valencian Government (CIT)
* 
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
* MA  02110-1301, USA.
* 
*/

/*
* Prodevelop, S.L.
* Pza. Juan de Villarrasa, 14 - 5
* Valencia
* Spain
* Tel: +34 963510612
* e-mail: prode@prodevelop.es
* http://www.prodevelop.es
*
*/
package es.prodevelop.gvsig.mini.namefinder;


import es.prodevelop.geodetic.utils.conversion.ConversionCoords;
import es.prodevelop.gvsig.mobile.fmap.proj.CRSFactory;
import es.prodevelop.gvsig.mini.utiles.Tags;
import es.prodevelop.gvsig.mini.utiles.Utilities;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.Vector;
import org.kxml2.io.KXmlParser;
import org.xmlpull.v1.XmlPullParserException;
import es.prodevelop.gvsig.mini.geom.Point;

/**
 * This class parses the response of the namefinder service
 * @author Alberto Romeu Carrasco (aromeu@prodevelop.es)
 */
public class NameFinder {

    /**
     * The result of the namefinder query
     */
    public Named[] searchResults;
    private static final String SEARCH_RESULTS = "searchresults";
    private static final String NAMED = "named";
    private static final String TYPE = "type";
    private static final String ID = "id";
    private static final String LAT = "lat";
    private static final String LON = "lon";
    private static final String NAME = "name";
    private static final String CATEGORY = "category";
    private static final String INFO = "info";
    private static final String RANK = "rank";
    private static final String IS_IN = "is_in";
    private static final String DISTANCE = "distance";
    private static final String APPROX_DISTANCE = "approxdistance";
    private static final String DIRECTION = "direction";
    private static final String ZOOM = "zoom";
    private static final String DESCRIPTION = "description";
    private static final String PLACE = "place";
    private static final String NEAREST_PLACES = "nearestplaces";
    private static final String ERROR = "error";
    public static final String URL = "http://gazetteer.openstreetmap.org/namefinder/search.xml?find=";
    public static final String MAX = "&max=";
    public boolean orderByRank = false;
    /**
     * The parameters of the query
     */
    public static String parms;

    /**
     * Parses the response of the namefinder service
     * @param data The data to be parsed
     */
    public Vector parse(final byte[] data) {
        final KXmlParser kxmlParser = new KXmlParser();
        Vector nameds = new Vector();
        try {
            kxmlParser.setInput(new ByteArrayInputStream(data), "UTF-8");
            kxmlParser.nextTag();
            int tag;
            if (kxmlParser.getEventType() != KXmlParser.END_DOCUMENT) {
                kxmlParser.require(KXmlParser.START_TAG, null,
                        SEARCH_RESULTS);
                tag = kxmlParser.nextTag();
                while (tag != KXmlParser.END_DOCUMENT) {                    
                    switch (tag) {
                        case KXmlParser.START_TAG:
                            if (kxmlParser.getName().compareTo(
                                    NAMED) == 0) {
                                nameds.addElement(this.parseNamed(kxmlParser));
                            } else if (kxmlParser.getName().compareTo(
                                    DISTANCE) == 0) {
                                //not supported
                            }
                            break;

                        case KXmlParser.END_TAG:
                            break;
                    }
                    tag = kxmlParser.next();
                }
            //kxmlParser.require(KXmlParser.END_DOCUMENT, null, null);
            }
        } catch (XmlPullParserException parser_ex) {
            parser_ex.printStackTrace();
        } catch (IOException ioe) {
            ioe.printStackTrace();
        } catch (OutOfMemoryError ou) {
            System.gc();
            System.gc();
        } finally {
            final int size = nameds.size();
            if (size > 0) {
                this.searchResults = new Named[size];
                nameds.copyInto(searchResults);
            }
            return nameds;
        }
    }

    /**
     * Parses a named tag 
     * @param parser
     * @return A Named instance
     */
    public Named parseNamed(final KXmlParser parser) {
        int currentTag;
        boolean end = false;
        Named named = new Named(0, 0);
        try {
            if (parser.getEventType() != KXmlParser.END_DOCUMENT) {
                parser.require(KXmlParser.START_TAG, null, NAMED);
            }
            named = this.parseNamedAttributes(parser, named);
            currentTag = parser.next();
            while (!end) {
                switch (currentTag) {
                    case KXmlParser.START_TAG:
                        if (parser.getName().compareTo(DESCRIPTION) == 0) {
                            final String desc = parser.nextText();
                            final String d = Utilities.parseNamedDescription(desc);
                            if (d == null || Utilities.isEmpty(d)) {
                                named.description = desc;
                            } else {
                                named.description = d;
                            }
                        } else if (parser.getName().compareTo(PLACE) == 0) {
                            named.place = parsePlace(parser);
                        } else if (parser.getName().compareTo(NEAREST_PLACES) == 0) {
                            named.nearestPlaces = parseNearestPlaces(parser);
                        }
                        break;
                    case KXmlParser.END_TAG:
                        if (parser.getName().compareTo(NAMED) == 0) {
                            end = true;
                        }
                        break;
//                    default:
//                        break;
                }
                if (!end) {
                    currentTag = parser.next();
                }
            }
        } catch (XmlPullParserException parser_ex) {
            parser_ex.printStackTrace();
        } catch (IOException ioe) {
            ioe.printStackTrace();
        } catch (OutOfMemoryError ou) {
            System.gc();
            System.gc();
        } finally {
            return named;
        }
    }

    /**
     * Parses the attributes of the named tag
     * @param parser
     * @param named The current named instance being parsed
     * @return The named parsed with its attributes filled
     */
    public Named parseNamedAttributes(final KXmlParser parser, Named named) {

        try {
            named.type = parser.getAttributeValue("", TYPE);

            final String id = parser.getAttributeValue("", ID);
            if (!Utilities.isEmpty(id)) {
                named.id = Integer.parseInt(id);
            }

            final String lat = parser.getAttributeValue("", LAT);

            final String lon = parser.getAttributeValue("", LON);
            if (!Utilities.isEmpty(lon) && !Utilities.isEmpty(lat)) {
                named.setLonLat(Double.parseDouble(lon), Double.parseDouble(lat), true);
            }

            named.name = parser.getAttributeValue("", NAME);
            named.category = parser.getAttributeValue("", CATEGORY);
            named.info = parser.getAttributeValue("", INFO);

            final String rank = parser.getAttributeValue("", RANK);
            if (!Utilities.isEmpty(rank)) {
                named.rank = Integer.parseInt(rank);
            }

            named.isIn = parser.getAttributeValue("", IS_IN);

            final String distance = parser.getAttributeValue("", DISTANCE);
            if (!Utilities.isEmpty(distance)) {
                named.distance = Double.parseDouble(distance);
            }

            final String approx = parser.getAttributeValue("", APPROX_DISTANCE);
            if (!Utilities.isEmpty(approx)) {
                named.approxDistance = Double.parseDouble(approx);
            }

            final String dir = parser.getAttributeValue("", DIRECTION);
            if (!Utilities.isEmpty(dir)) {
                named.directionAngle = Double.parseDouble(dir);
            }

            final String zoom = parser.getAttributeValue("", ZOOM);
            if (!Utilities.isEmpty(zoom)) {
                named.zoom = Integer.parseInt(zoom);
            }

//        } catch (XmlPullParserException parser_ex) {
//            parser_ex.printStackTrace();
//        } catch (IOException ioe) {
//            ioe.printStackTrace();
        } catch (Exception e) {
            e.printStackTrace();
        } catch (OutOfMemoryError ou) {
            System.gc();
            System.gc();
        } finally {
            return named;
        }
    }

    /**
     * Parses an error response
     * @param parser
     * @return True if error was found
     */
    public boolean parseError(final KXmlParser parser) {
        boolean error = false;
        try {
            parser.next();
            if (parser.getAttributeValue("", ERROR) != null) {
                error = true;
            }

//        } catch  (XmlPullParserException parser_ex) {
//            parser_ex.printStackTrace();
//        } catch (IOException ioe) {
//            ioe.printStackTrace();
        } catch (OutOfMemoryError ou) {
            System.gc();
            System.gc();
        } finally {
            return error;
        }
    }

    /**
     * Parses the place tag
     * @param parser
     * @return The named instance of the place tag
     */
    public Named parsePlace(final KXmlParser parser) {
        int currentTag;
        boolean end = false;
        Named named = new Named(0, 0);
        try {
            if (parser.getEventType() != KXmlParser.END_DOCUMENT) {
                parser.require(KXmlParser.START_TAG, null, PLACE);
            }
            currentTag = parser.next();
            while (!end) {
                switch (currentTag) {
                    case KXmlParser.START_TAG:
                        if (parser.getName().compareTo(NAMED) == 0) {
                            named = this.parseNamed(parser);
                        }
                        break;
                    case KXmlParser.END_TAG:
                        if (parser.getName().compareTo(PLACE) == 0) {
                            end = true;
                        }
                        break;
                    case KXmlParser.TEXT:
                        break;
                    default:
                        break;
                }
                if (!end) {
                    currentTag = parser.next();
                }
            }
        } catch (XmlPullParserException parser_ex) {
            parser_ex.printStackTrace();
        } catch (IOException ioe) {
            ioe.printStackTrace();
        } catch (OutOfMemoryError ou) {
            System.gc();
            System.gc();
        } finally {
            return named;
        }
    }

    /**
     * Parses the nearestplaces tag
     * @param parser
     * @return An array of named instances
     */
    public Named[] parseNearestPlaces(final KXmlParser parser) {
        int currentTag;
        boolean end = false;
        Vector places = new Vector();
        Named[] results = null;
        try {
            if (parser.getEventType() != KXmlParser.END_DOCUMENT) {
                parser.require(KXmlParser.START_TAG, null, NEAREST_PLACES);
            }

            currentTag = parser.next();
            while (!end) {
                switch (currentTag) {
                    case KXmlParser.START_TAG:
                        if (parser.getName().compareTo(NAMED) == 0) {
                            places.addElement(this.parseNamed(parser));
                        }
                        break;
                    case KXmlParser.END_TAG:
                        if (parser.getName().compareTo(NEAREST_PLACES) == 0) {
                            end = true;
                        }
                        break;
                    case KXmlParser.TEXT:
                        break;
                    default:
                        break;
                }
                if (!end) {
                    currentTag = parser.next();
                }
            }
        } catch (XmlPullParserException parser_ex) {
            parser_ex.printStackTrace();
        } catch (IOException ioe) {
            ioe.printStackTrace();
        } catch (OutOfMemoryError ou) {
            System.gc();
            System.gc();
        } finally {
            final int size = places.size();
            if (size > 0) {
                results = new Named[size];
                places.copyInto(results);
            }
            return results;
        }
    }
   
    /**
     * Orders the results by its rank from major to minor rank
     * @param a The array of Named to be ordered
     * @param center The rank to compare with
     */
    public static void quicksort(final Named[] a) {
        quicksort(a, 0, a.length - 1);
    }

    // quicksort a[left] to a[right]
    private static void quicksort(final Named[] a, final int left, final int right) {
        if (right <= left) {
            return;
        }
        int i = partition(a, left, right);
        quicksort(a, left, i - 1);
        quicksort(a, i + 1, right);
    }

    // partition a[left] to a[right], assumes left < right
    private static int partition(final Named[] a, final int left, final int right) {
        int i = left - 1;
        int j = right;
        while (true) {

            while (less(a[right].rank, a[++i].rank)) // find item on left to swap
            ;                               // a[right] acts as sentinel

            while (less(a[right].rank, a[right].rank)) // find item on right to swap
            {
                if (j == left) {
                    break;           // don't go out-of-bounds

                }
            }
            if (i >= j) {
                break;                  // check if pointers cross

            }
            exch(a, i, j);                      // swap two elements into place

        }
        exch(a, i, right);                      // swap with partition element

        return i;
    }

    // is x < y ?
    private static boolean less(double x, double y) {
        return (x < y);
    }

    // exchange a[i] and a[j]
    private static void exch(Named[] a, int i, int j) {
        Named swap = a[i];

        a[i] = a[j];
        a[j] = swap;
    }

    private static boolean isNearCity(final Named nam, final String nearCity) {
        if (nearCity == null) {
            return false;
        }
        
        final Named[] nearest = nam.nearestPlaces;
        if (nearest != null) {
            final int size = nearest.length;
            Named n;
            for (int i = 0; i < size; i++) {
                n = nearest[i];
                if (n.name.toLowerCase().compareTo(nearCity.trim().toLowerCase()) == 0) {                    
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * Sets first the results of the namefinder service if they are near from 
     * the city specified. Being near from the city means that contains a Named inside
     * the NearestPlaces tag that contains the name specified in city
     * @param res The results of the namefinder service
     * @param city The city to search
     * @return The results ordered
     */
    public Named[] orderNearCity(Named[] res, final String city) {
        final int size = res.length;
        final Named[] result = new Named[size];
        
        int left = 0;
        int right = size-1;
        for(int i = 0; i<size; i++) {
            Named n = res[i];
            if(this.isNearCity(n, city)) {
                result[left] = n;
                left++;
            } else {
                result[right] = n;
                right--;
            }
        }
        return result;
    }
}
