/* SIGATEX. Gestión de Recursos Turísticos de Extremadura
  * Funded by European Union. FEDER program.
  *
  * Copyright (C) 2008 Junta de Extremadura.
  * Developed by: Prodevelop, S.L.
  * For more information, contact:
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation, either version 2 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-  1307,USA.
  *
  * For more information, contact:
  *
  * Junta de Extremadura
  * Consejería de Cultura y Turismo
  * C/ Almendralejo 14 Mérida
  * 06800 Badajoz
  * SPAIN
  *
  * Tel: +34 924007009
  * http://www.culturaextremadura.com
  *
  * or
  *
  * Prodevelop, S.L.
  * Pza. Juan de Villarrasa, 14 - 5
  * Valencia
  * Spain
  * Tel: +34 963510612
  * e-mail: prode@prodevelop.es
  * http://www.prodevelop.es
  *
  */
package es.prodevelop.gvsig.mini.utiles;

import java.util.Vector;

/**
 * A Thread Pool implementation to execute tasks in background
 * @author Alberto Romeu Carrasco (aromeu@prodevelop.es)
 * @author aromeu - modified for gvSIG phone
 */
public class ThreadPool {

    /**
     * The backlog of assignments, which are waiting
     * for the thread pool.
     */
    protected final Vector assignments = new Vector();
    private final Object lock = this;
    private boolean disposed = false;
    private static ThreadPool instance = null;
    /**
     * We initialize this instance with only one thread, to execute tasks sequentially
     */
    private static ThreadPool instance2 = null;

    /**
     * This constructor initializes the threadpool with the number of threads specified
     * @param size The number of threads of the threadpool
     */
    protected ThreadPool(int size) {
        WorkerThread thread;
        for (int i = 0; i < size; i++) {
            thread = new WorkerThread();
            thread.setPriority(Thread.MIN_PRIORITY);
            thread.start();
        }
    }

    /**
     * Initializes a threadpool instance with a single thread
     */
    protected ThreadPool() {
        WorkerThread thread = new WorkerThread();
        thread.setPriority(Thread.MIN_PRIORITY);
        thread.start();
    }

    /**
     * Singleton
     * @return instance
     */
    public static ThreadPool getInstance(int size) {
        if (instance == null) {
            instance = new ThreadPool(size);
        }
        return instance;
    }

    /**
     * This method return a ThreadPool with only one Thread. Use this instance
     * for tasks which can be executed sequentially or single threads.
     * @return
     */
    public static ThreadPool getInstance() {
        if (instance2 == null) {
            instance2 = new ThreadPool();
        }
        return instance2;
    }

    /**
     * Disposes of TileThreadPool. No new jobs can be added, and worker 
     * threads will terminate once the current job queue is drained.
     */
    public void dispose() {
        synchronized (lock) {
            disposed = true;
            lock.notifyAll();
        }
    }

    /**
     * Add a task to the thread pool. Any class
     * which implements the Runnable interface
     * may be assienged. When this task runs, its
     * run method will be called.
     *
     * @param task   An object that implements the Runnable interface
     */
    public void assign(Runnable task) {
        if (task == null) {
            throw new IllegalArgumentException("Null task");
        }
        synchronized (lock) {
            if (disposed) {
                throw new IllegalStateException("Attempt to add job to disposed ThreadPool queue.");
            }
            if (!assignments.contains(task)) {
                assignments.addElement(task);
                lock.notify();
            }
        }
    }
    
    /**
     * Add a task to the thread pool in the first position of the queue.
     * Any class which implements the Runnable interface may be assienged. 
     *
     * @param task   An object that implements the Runnable interface
     */
    public void assignFirst(Runnable task) {
        if (task == null) {
            throw new IllegalArgumentException("Null task");
        }
        synchronized (lock) {
            if (disposed) {
                throw new IllegalStateException("Attempt to add job to disposed ThreadPool queue.");
            }
            if (!assignments.contains(task)) {
                assignments.insertElementAt(task, 0);
                lock.notify();
            }
        }
    }

    /*
     * Cancel all the assignments
     * */
    public void clearAll() {
        synchronized (lock) {
            assignments.removeAllElements();
            lock.notify();
        }
    }

    private class WorkerThread extends Thread {

        WorkerThread() {
            super();
        }

        public void run() {
            while (true) {
                // Get a job from the queue
                Runnable job = null;
                synchronized (lock) {
                    //while (job == null) {
                        if (assignments.size() > 0) {
                            job = (Runnable) assignments.firstElement();
                            assignments.removeElementAt(0);
                        } else if (disposed) {
                            return;
                        } else {
                            try {
                                lock.wait();
                            } catch (InterruptedException e) {
                                e.printStackTrace();
                            //Logger.info("InterruptedException "+this.getName());
                            }
                        }
                    //}
                }
                // Run it
                boolean done = false;
                try {
                    if(job != null)
                    job.run();
                    done = true;
                    job = null;
                } finally {
                    if (!done) {
                        // Some kind of exception/error. Allow it to be caught by the default
                        // exception handler, but start another worker to replace this one.
                        new WorkerThread().start();
                    }
                }
            }
        }
    }
}
